"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.mockObject = void 0;
const mockFn_js_1 = require("./mockFn.js");
/**
 * Returns a mock object that can be used in place of a real object. This is
 * useful if the real object is hard to create or if it is hard to get it
 * exhibit some specific behavior.
 *
 * Functions on the mock object are also `mockFn`s, so they can be used in
 * validators that work for mocks.
 *
 * @param overrides - The members to be overridden.
 *
 * @example
 * ```ts
 * class Person {
 *   constructor(public name: string, public age: number) {}
 *
 *   isAdult() {
 *     return this.age >= 18
 *   }
 * }
 *
 * const mock = mockObject<Person>({
 *   isAdult: () => true,
 * })
 *
 * expect(mock.isAdult()).toEqual(true)
 * expect(mock.isAdult).toHaveBeenCalled()
 * ```
 */
function mockObject(overrides = {}) {
    const clone = replaceFunctionsWithMocks(overrides);
    return new Proxy(clone, {
        get(target, property, receiver) {
            if (Reflect.has(target, property)) {
                return Reflect.get(target, property, receiver);
            }
            throw new TypeError(`Cannot access .${property.toString()} - no mock value provided.`);
        },
    });
}
exports.mockObject = mockObject;
function replaceFunctionsWithMocks(object) {
    const clone = { ...object };
    for (const key of Object.keys(clone)) {
        const value = clone[key];
        if (typeof value === 'function') {
            if (!(0, mockFn_js_1.isMockFn)(value)) {
                clone[key] = (0, mockFn_js_1.mockFn)(value);
            }
        }
    }
    return clone;
}
/**
 * Resets every `mockFn` of the given `MockObject`.
 *
 * @param mockObject - The `MockObject` to reset
 */
mockObject.reset = function reset(mockObject) {
    for (const value of Object.values(mockObject)) {
        if ((0, mockFn_js_1.isMockFn)(value)) {
            value.reset();
        }
    }
};
//# sourceMappingURL=mockObject.js.map